/***********************************************************
	msh - menu shell

	0.00 85-08-15 ldab/fl: first version.
	1.00 85-10-14 ldab/hn: Can handle more than one menu
	2.00 85-11- ldab/hn: It's a shell now
	by Henrik Norbeck, Luxor Datorer AB 

	menuread.c Contains menu initialization functions

	getmenu  - reads a menu
	getcolon - gets an input line from the structure file
		   and splits it at the colons.
	split	 - splits a line into substrings separated 
		   by a certain separator character
	getline  - reads a line from a file
	noescape - deletes 'escape' characters in a string

***********************************************************/

/* Header files */
#include <stdio.h>
#include "msh.h"

/* user information storage */
extern char *homedir;	/* home directory */
extern char userid[];	/* user identification */
extern char grpid[];
extern char systemid[];	/* system identification */
extern char *g_program;	/* global prog name */
extern char structfile[];

/*	string arrays for the menus */
extern char mbuf[MENUSIZE][MENULINE];
extern char pbuf[MENUSIZE][PROGLINE];
extern char nbuf[MENUSIZE][NAMELINE];

extern char menu_name[];

/*** Getmenu() reads the menu called menu_name. If menu_name=NULL
	it reads the first menu in the structure file. */
getmenu()
{
    char *noescape();
    char *str[5];				/* operative pointers */
    char buf[256];				/* file string buffer */
    FILE *fp, *fopen();				/* file structures */
    int found, eof;
    register int i, n, menu, prog;
		
    /* phase 1: load menu structure for current page */
    if ((fp = fopen(structfile, "r")) == NULL)
	panic("no structure file");

    eof = (getcolon(fp, buf, str) == NULL);
    do {
	n = menu = prog = 0;
	while (!eof) {
	    if (*str[STR_PROG] != '\0') /* programs */
		strncpy(pbuf[n], str[STR_PROG], PROGLINE);
	    else {
		prog--;
		*pbuf[n] = MARKCHAR;
		strcpy(pbuf[n]+1, str[STR_SYMB]);
		}
	    if (*str[STR_TEXT] != '\0') /* commands */
		strncpy(mbuf[n], noescape(str[STR_TEXT]), MENULINE);
	    else {
		menu--;
		*mbuf[n] = MARKCHAR;
		strcpy(mbuf[n]+1, str[STR_SYMB]);
		}
	    strncpy(nbuf[n], noescape(str[STR_NAME]), NAMELINE);
	    if (*str[STR_NAME] == '$' && n != 0)
		break;
	    n++;
	    eof = (getcolon(fp, buf, str) == NULL);
	    }
	if (*nbuf[0] != '$')
	    panic("syntax error in structure file");
	if (found = (*menu_name == '\0'))
	    break;	/* Should be if (menu_name == NULL) */
    } while (!(found = (strncmp(menu_name, nbuf[0], NAMELINE) == 0)) && !eof);
    if (!found)
	panic("menu not found in structure file");
    fclose(fp);	/* close */

    /* phase 2: try to solve open links */

    if (menu || prog) {
	if ((fp = fopen(CATALOGUE, "r")) == NULL)
	    panic("no global catalogue");
	while ((menu < 0 || prog < 0) && getcolon(fp,buf,str) != NULL) {
	    for (i = n;--i >= 0;)
		if (*pbuf[i] == MARKCHAR
			&& strcmp(str[CAT_SYMB], pbuf[i] + 1) == 0) {
		    prog++;
		    strncpy(pbuf[i], str[CAT_PROG], PROGLINE );
		    break;
		    }
	    for (i = n;--i >= 0;)
		if (*mbuf[i] == MARKCHAR
			&& strcmp(str[CAT_SYMB], mbuf[i] + 1) == 0) {
		    menu++;
		    strncpy(mbuf[i], noescape(str[CAT_TEXT]), MENULINE);
		    break;
		    }
	    }
	fclose(fp);
    }

    /* phase 3: mark all unsolved entries */

    if (prog || menu)
	for (i = n; --i >= 0;) {
	    if (*pbuf[i] == MARKCHAR)
		strncpy(pbuf[i], "", PROGLINE);
	    if (*mbuf[i] == MARKCHAR)
		strncpy(mbuf[i], "---", MENULINE);
	    }
    return(n);
}

/***	get input lines

	get a line from fp and split it into a number of
 	substrings, separated by colon. the function returns
	a pointer to a array of character pointers, or NULL
	if failure. the function does not handle overflow.
*/
getcolon(fp, buf, str) FILE *fp; char *buf, **str;
{
    char *getline();

    if (getline(fp, buf) == NULL)
	return NULL;

    split(5, buf, ':', str);
    return NOTNULL;
}

/***	split input string

	Splits the input string into a number of substrings
 	separated by a given char. The function returns a
	pointer to an array of character pointers. No over-
	flow test.

	To type a separator character in a string, an 'escape'
	character is used as a prefix. That character are
	ignored, and any character following is passed by.
	Escape chars are not deleted from the split strings.

	Note: The split is made by placing null separators
	in the input string, thus destroying the original
	contents.
*/
split(cnt, buf, chr, str)
register short int cnt;
char *buf, chr, **str;
{
    register char *p = buf, *c = buf;
    register short int n = 0;

    while (*p && n < cnt) {
	if (*p != '\\') {		/* ignore escape */
	    if (*p == chr) {		/* break on character */
		str[n++] = c;		/* save the pointer */
		*p = '\0';		/* terminate string */
		c = p + 1;		/* set pointer */
		}
	    }
	else
	    p++;
	p++;				/* next character */
	}
    str[n++] = c;
    while (n < cnt) str[n++] = p;	/* fix nulls */
}

/*** Getline() reads one line from a file, skipping comment lines */
char *getline(fp, buffer)
FILE *fp;
char *buffer;
{
    register char *p;
    register char c;

    while ((c=getc(fp)) == '#' || c == '\n')	/* Skip comment lines */
	if (c == '#')				/* and empty lines */
	    while ((c=getc(fp)) != EOF && c != '\n');

    p = buffer;
    while (c != EOF) {
	if (c == '\n')
	    if (*(p-1) != '\\')
		break;
	    else
		p--;	/* 'Escaped' return! Skip it! */
	else
	    *p++ = c;
	c = getc(fp);
	}
    if (c == EOF) return (NULL);
    *p = '\0';
    return (buffer);
}

/***	delete 'escape' characters */
char *noescape(ptr)
register char *ptr;
{
    register char *p = ptr, *s = ptr;
    do {
	if (*p == '\\') p++;
    } while (*ptr++ = *p++);
    return (s);
}

/***    Take away escapes from a string and surrounding quotes ***/
char *noescquot(s)
register char *s;
{
    char *noescape();

    if (*s == '"' || *s == '\'') {
        strcpy(s, s+1);
        s[strlen(s)-1] = '\0';
        }
    return (noescape(s));
}
